import json
import time
from dataclasses import dataclass, asdict
from typing import Any
import requests
from websockets.sync.client import connect, ClientConnection
from dogu.device.appium_server import AppiumContextServerInfo, AppiumServerContext
from dogu.device.common.device_closer import DeviceCloser
from dogu.device.common.device_http_response import DeviceHttpResponse
from dogu.device.common.const import (
    DOGU_DEVICE_AUTHORIZATION_HEADER_KEY,
    DOGU_DEVICE_SERIAL_HEADER_KEY,
)
from dogu.device.logger import create_logger


@dataclass(frozen=True)
class DeviceForwardSendMessage:
    # pylint: disable=invalid-name
    serial: str
    # pylint: disable=invalid-name
    hostPort: int
    # pylint: disable=invalid-name
    devicePort: int


@dataclass(frozen=True)
class DeviceRunAppiumServerSendMessage:
    # pylint: disable=invalid-name
    serial: str


@dataclass(frozen=True)
class AppiumContextInfo:
    # pylint: disable=invalid-name
    serial: str
    # pylint: disable=invalid-name
    server: AppiumContextServerInfo


@dataclass(frozen=True)
class GetAppiumContextInfoResponse:
    info: AppiumContextInfo


class DeviceClient:
    def __init__(self, device_server_url: str, token: str, timeout: int):
        self.device_server_url = (
            device_server_url
            if device_server_url[-1] != "/"
            else device_server_url[:-1]
        )
        self._token = token
        self.timeout = timeout
        self._logger = create_logger(__name__)

    def forward(self, serial: str, host_port: int, device_port: int) -> DeviceCloser:
        conn = self.__subscribe("/ws/devices/forward", serial=serial)
        param = DeviceForwardSendMessage(
            serial=serial, hostPort=host_port, devicePort=device_port
        )
        json_param = json.dumps(asdict(param))
        conn.send(json_param)
        while True:
            msg = conn.recv()
            msg_json = json.loads(msg)
            kind = msg_json["value"]["kind"]
            if kind != "DeviceForwardReceiveMessageResultValue":
                continue
            success = msg_json["value"]["success"]
            if not success:
                raise Exception(
                    f"Failed to forward. error: {msg_json['value']['reason']}"
                )
            break
        return DeviceCloser(conn)

    def run_appium_server(self, serial: str) -> AppiumServerContext:
        conn = self.__subscribe("/ws/devices/run-appium-server", serial=serial)
        param = DeviceRunAppiumServerSendMessage(serial=serial)
        json_param = json.dumps(asdict(param))
        conn.send(json_param)
        while True:
            msg = conn.recv()
            msg_json = json.loads(msg)
            kind = msg_json["value"]["kind"]
            if kind != "DeviceRunAppiumServerReceiveMessageResultValue":
                continue
            success = msg_json["value"]["success"]
            if not success:
                raise Exception(
                    f"Failed to run_appium_server. error: {msg_json['value']['reason']}"
                )
            break
        return AppiumServerContext(
            AppiumContextServerInfo(port=int(msg_json["value"]["serverPort"])), conn
        )

    def get_appium_capabilities(self, serial: str) -> Any:
        full_path = f"{self.device_server_url}/devices/{serial}/appium-capabilities"
        headers = {}
        headers[DOGU_DEVICE_AUTHORIZATION_HEADER_KEY] = self._token

        res = requests.get(full_path, timeout=self.timeout, headers=headers)
        res.raise_for_status()
        device_res = DeviceHttpResponse(res)
        if device_res.error()[0]:
            raise Exception(
                f"DeviceClient.get_appium_capabilities error: {device_res.error()[1].message}"
            )
        return device_res.data_dict()["capabilities"]

    def __subscribe(
        self, path: str, serial: str, try_count: int = 5
    ) -> ClientConnection:
        ws_url = self.device_server_url.replace("http://", "ws://").replace(
            "https://", "wss://"
        )
        full_path = f"{ws_url}{path}"

        headers = {}
        headers[DOGU_DEVICE_AUTHORIZATION_HEADER_KEY] = self._token
        headers[DOGU_DEVICE_SERIAL_HEADER_KEY] = serial

        last_error = None
        for i in range(try_count):
            try:
                socket = connect(
                    full_path,
                    additional_headers=headers,
                )
                return socket
            except Exception as error:
                last_error = error
                self._logger.info(
                    "Failed to connect to %s. count: (%d/%d), error: %s",
                    full_path,
                    i + 1,
                    try_count,
                    error,
                )
                time.sleep(1)
                continue
        raise Exception(f"Failed to connect to {full_path}. error: {last_error}")



from typing import Any, TypeVar, TypedDict
from dacite import from_dict

from requests import Response

T = TypeVar("T")


class DeviceError(TypedDict):
    code: int
    message: str


class DeviceHttpResponse:
    def __init__(self, res: Response):
        self.res_json = res.json()

    def error(self) -> [bool, DeviceError]:
        if self.res_json["value"]["$case"] != "error":
            return False, DeviceError(code=0, message="")

        return True, DeviceError(
            code=self.res_json["value"]["error"]["code"],
            message=self.res_json["value"]["error"]["message"],
        )

    def data(self, cls: T) -> T:
        data_dict = self.data_dict()
        return from_dict(data_class=cls, data=data_dict)

    def data_dict(self) -> Any:
        if self.res_json["value"]["$case"] != "data":
            raise Exception("DeviceHttpResponse not a data type")
        return self.res_json["value"]["data"]



from dataclasses import asdict, dataclass
from typing import List, Optional
import requests
from dogu.device.common.device_http_response import DeviceHttpResponse
from dogu.device.logger import create_logger
from dogu.device.common.const import DOGU_DEVICE_AUTHORIZATION_HEADER_KEY


@dataclass(frozen=True)
class GetFreePortQuery:
    # pylint: disable=invalid-name
    excludes: List[int]
    # pylint: disable=invalid-name
    offset: int


@dataclass(frozen=True)
class GetFreePortResponse:
    # pylint: disable=invalid-name
    port: int


@dataclass(frozen=True)
class EnsureBrowserAndDriverOptions:
    # pylint: disable=invalid-name
    browserName: str

    # pylint: disable=invalid-name
    browserPlatform: str

    # pylint: disable=invalid-name
    browserVersion: Optional[str]

    # pylint: disable=invalid-name
    deviceSerial: Optional[str]


# pylint: disable=too-many-instance-attributes
@dataclass(frozen=True)
class EnsureBrowserAndDriverResult:
    # pylint: disable=invalid-name
    browserName: str

    # pylint: disable=invalid-name
    browserPlatform: str

    # pylint: disable=invalid-name
    browserVersion: str

    # pylint: disable=invalid-name
    browserMajorVersion: int

    # pylint: disable=invalid-name
    browserDriverVersion: str

    # pylint: disable=invalid-name
    browserDriverPath: str

    # pylint: disable=invalid-name
    browserPath: Optional[str]

    # pylint: disable=invalid-name
    browserPackageName: Optional[str]

    # pylint: disable=invalid-name
    deviceSerial: Optional[str]


class DeviceHostClient:
    def __init__(self, device_server_url: str, token: str, timeout: int = 10):
        self.device_server_url = (
            device_server_url
            if device_server_url[-1] != "/"
            else device_server_url[:-1]
        )
        self._token = token
        self.timeout = timeout
        self._logger = create_logger(__name__)

    def get_free_port(self, excludes: Optional[List[int]] = None) -> int:
        if not excludes:
            excludes = []
        full_path = f"{self.device_server_url}/device-host/free-port"
        param = GetFreePortQuery(excludes=excludes, offset=0)
        headers = {}
        headers[DOGU_DEVICE_AUTHORIZATION_HEADER_KEY] = self._token

        res = requests.get(
            full_path, json=asdict(param), timeout=self.timeout, headers=headers
        )
        res.raise_for_status()
        device_res = DeviceHttpResponse(res)
        if device_res.error()[0]:
            raise Exception(
                f"DeviceHostClient.get_free_port error: {device_res.error()[1].message}"
            )
        res_obj = device_res.data(GetFreePortResponse)
        return res_obj.port

    def ensure_browser_and_driver(
        self, options: EnsureBrowserAndDriverOptions
    ) -> EnsureBrowserAndDriverResult:
        full_path = f"{self.device_server_url}/device-host/ensure-browser-and-driver"
        headers = {}
        headers[DOGU_DEVICE_AUTHORIZATION_HEADER_KEY] = self._token

        res = requests.post(
            full_path, json=asdict(options), timeout=self.timeout, headers=headers
        )
        res.raise_for_status()
        device_res = DeviceHttpResponse(res)
        if device_res.error()[0]:
            raise Exception(
                f"DeviceHostClient.ensure_browser_and_driver error: {device_res.error()[1].message}"
            )
        res_obj = device_res.data(EnsureBrowserAndDriverResult)
        return res_obj



from websockets.sync.client import ClientConnection
from dogu.device.common.closable import IClosable


class DeviceCloser(IClosable):
    def __init__(self, conn: ClientConnection):
        self.conn = conn

    def close(self) -> None:
        self.conn.close()



from dataclasses import dataclass
from websockets.sync.client import ClientConnection

from dogu.device.common.device_closer import DeviceCloser


@dataclass(frozen=True)
class AppiumContextServerInfo:
    # pylint: disable=invalid-name
    port: int


class AppiumServerContext:
    def __init__(self, info: AppiumContextServerInfo, conn: ClientConnection):
        self.info = info
        self.closer = DeviceCloser(conn)

    @property
    def port(self) -> int:
        return self.info.port

    def close(self) -> None:
        self.closer.close()



import logging


def create_logger(name: str):
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger
